class TaskDefinitionError(Exception):
    """
    Raised when there is an error in the definition of the task list used by
    ``jogger``.
    """
    
    pass


class TaskError(Exception):
    """
    Used to indicate problem during the execution of a task, yielding a nicely
    printed error message in the appropriate output stream (e.g. ``stderr``).
    """
    
    pass



import os
from collections import OrderedDict

from jogger.utils.files import walk

from .base import Task, TaskError

try:
    import flake8  # noqa
    HAS_FLAKE8 = True
except ImportError:
    HAS_FLAKE8 = False

try:
    import isort  # noqa
    HAS_ISORT = True
except ImportError:
    HAS_ISORT = False

try:
    import bandit  # noqa
    HAS_BANDIT = True
except ImportError:
    HAS_BANDIT = False

try:
    from django.core.management import call_command  # noqa
    HAS_DJANGO = True
except ImportError:
    HAS_DJANGO = False

ENDINGS = {
    'CRLF': b'\r\n',
    'CR': b'\r',
    'LF': b'\n'
}

DEFAULT_GOOD_ENDING = 'LF'
DEFAULT_MAX_FILESIZE = 1024 * 1024  # 1MB in bytes


def listify_multiline_string(string):
    """
    Return a list constructed by splitting the given multiline string,
    stripping whitespace, and filtering out empty values.
    
    :param string: The multiline string to convert into a list.
    :return: The resulting list.
    """
    
    result = [i.strip() for i in string.splitlines()]
    return filter(None, result)


class LintTask(Task):
    
    help = (
        'Lint the project. Automatically detects, and uses if found, isort and '
        'flake8 for linting Python code, and bandit for finding common security '
        'issues in Python code. Also runs fable (Find All Bad Line Endings) and '
        'performs a dry-run of makemigrations (if Django is detected).'
    )
    
    steps = [
        ('python', '-p', 'Perform linting of Python code.'),
        ('fable', '-f', 'Find all bad line endings.'),
        ('bandit', '-b', 'Perform a Bandit security scan.'),
        ('migrations', '-m', 'Perform makemigrations dry-run.')
    ]
    
    def __init__(self, *args, **kwargs):
        
        super().__init__(*args, **kwargs)
        
        self.outcomes = OrderedDict()
    
    def add_arguments(self, parser):
        
        for name, short_flag, help_text in self.steps:
            parser.add_argument(
                short_flag, f'--{name}',
                action='store_true',
                dest=f'do_{name}',
                help=help_text
            )
    
    def handle(self, **options):
        
        settings = self.settings
        
        explicit_steps = []
        implicit_steps = []
        for step, _, _ in self.steps:
            # Consider the step explicit if a command line flag is included for
            # it. Otherwise, consider the step implicit unless it has been
            # disabled via the project settings file
            if options[f'do_{step}']:
                explicit_steps.append(step)
            elif settings.getboolean(step, fallback=True):
                implicit_steps.append(step)
        
        if explicit_steps:
            run = explicit_steps
            explicit = True
        else:
            run = implicit_steps
            explicit = False
        
        for step in run:
            getattr(self, f'handle_{step}')(explicit)
        
        summary = []
        for label, result in self.outcomes.items():
            if result:
                styled_result = self.styler.success('OK')
            else:
                styled_result = self.styler.error('FAIL')
            
            summary.append(f'{label}: {styled_result}')
        
        if summary:
            self.stdout.write('Summary', style='label')
            self.stdout.write('\n'.join(summary))
    
    def handle_python(self, explicit):
        
        if explicit and not HAS_ISORT and not HAS_FLAKE8:
            self.stderr.write('Cannot lint python: Neither isort nor flake8 are available.')
            return
        
        if HAS_ISORT:
            self.stdout.write('Running isort...', style='label')
            result = self.cli('isort --check-only --diff .')
            self.outcomes['isort'] = result.returncode == 0
            self.stdout.write('')  # newline
        
        if HAS_FLAKE8:
            self.stdout.write('Running flake8...', style='label')
            result = self.cli('flake8 .')
            self.outcomes['flake8'] = result.returncode == 0
            self.stdout.write('')  # newline
    
    def _get_fable_excludes(self):
        
        # Start with some sane default exclusions
        excludes = {'.git', '__pycache__', '*.pyc', '*.pdf', '*.png', '*.jpg', '*.jpeg', '*.gif'}
        
        # Add any configured excludes
        try:
            extra_excludes = self.settings['fable_exclude']
        except KeyError:
            pass
        else:
            excludes.update(listify_multiline_string(extra_excludes))
        
        return excludes
    
    def handle_fable(self, explicit):
        
        self.stdout.write('Running fable...', style='label')
        
        excludes = self._get_fable_excludes()
        
        # Get the appropriate good ending from settings
        good_ending = self.settings.get('fable_good_endings', DEFAULT_GOOD_ENDING)
        if good_ending not in ENDINGS:
            raise TaskError(f'Invalid value for fable_good_endings setting ({good_ending}).')
        
        # Compile inverse dictionary of bad line endings
        bad_endings = {v: k for k, v in ENDINGS.items() if k != good_ending}
        
        # Get the maximum file size to analyse from settings
        max_filesize = self.settings.get('fable_max_filesize', DEFAULT_MAX_FILESIZE)
        
        try:
            max_filesize = int(max_filesize)
        except ValueError:
            raise TaskError(f'Invalid value for fable_max_filesize setting ({max_filesize}).')
        
        result = True
        skipped = 0
        for filename in walk('./', excludes):
            if os.path.getsize(filename) > max_filesize:
                skipped += 1
                continue
            
            with open(filename, 'rb') as f:
                content = f.read()
                for ending in bad_endings:
                    if ending in content:
                        self.stdout.write(f'Detected {bad_endings[ending]}: {filename}')
                        result = False
                        break
        
        if skipped:
            self.stdout.write(f'Skipped {skipped} large files')
        
        self.outcomes['fable'] = result
        self.stdout.write('')  # newline
    
    def handle_bandit(self, explicit):
        
        if not HAS_BANDIT:
            if explicit:
                self.stderr.write('Cannot run bandit: Package is not available.')
            
            return
        
        self.stdout.write('Running bandit...', style='label')
        
        cmd = 'bandit . -r'
        
        # Set the command's verbosity based on the verbosity level of the task
        verbosity = self.kwargs['verbosity']
        if verbosity < 2:
            cmd = f'{cmd} -q'  # run in "quiet" mode
        elif verbosity > 2:
            cmd = f'{cmd} -v'  # run in "verbose" mode
        
        # Add any configured excludes
        try:
            excludes = self.settings['bandit_exclude']
        except KeyError:
            pass
        else:
            excludes = ','.join(listify_multiline_string(excludes))
            cmd = f'{cmd} -x {excludes}'
        
        result = self.cli(cmd)
        self.outcomes['bandit'] = result.returncode == 0
        self.stdout.write('')  # newline
    
    def handle_migrations(self, explicit):
        
        if explicit and not HAS_DJANGO:
            self.stderr.write('Cannot check migrations: Django is not available.')
            return
        
        if HAS_DJANGO:
            self.stdout.write('Checking for missing migrations...', style='label')
            
            result = self.cli('python manage.py makemigrations --dry-run --check')
            
            self.outcomes['migrations'] = result.returncode == 0
            self.stdout.write('')  # newline



import os

from jogger.utils.input import JOG_FILE_NAME, find_config_file

from .base import Task, TaskError

try:
    import sphinx  # noqa
    HAS_SPHINX = True
except ImportError:
    HAS_SPHINX = False


class DocsTask(Task):
    
    help = (
        'Build the project documentation using Sphinx.'
    )
    
    def add_arguments(self, parser):
        
        parser.add_argument(
            '-f', '--full',
            action='store_true',
            dest='full',
            help=(
                'Remove previously built documentation before rebuilding all '
                'pages from scratch.'
            )
        )
        
        parser.add_argument(
            '-l', '--link',
            action='store_true',
            dest='link_only',
            help='Output the link to previously built documentation and exit.'
        )
    
    def handle(self, **options):
        
        if not HAS_SPHINX:
            raise TaskError('Sphinx not detected.')
        
        # Get the project directory by locating the task file (jog.py). Assume
        # a "docs" directory under that.
        project_dir = os.path.dirname(find_config_file(JOG_FILE_NAME))
        docs_dir = os.path.join(project_dir, 'docs')
        
        if not os.path.exists(docs_dir):
            raise TaskError(f'Documentation directory not found at {docs_dir}.')
        
        if options['link_only']:
            show_link = True
            output_prefix = ''
        else:
            command = [f'cd {docs_dir}']
            if options['full']:
                command.append('make clean')
            
            command.append('make html')
            
            result = self.cli(' && '.join(command))
            show_link = result.returncode == 0
            output_prefix = '\n'
        
        if show_link:
            index_path = os.path.join(docs_dir, '_build', 'html', 'index.html')
            if os.path.exists(index_path):
                self.stdout.write(self.styler.label(
                    f'{output_prefix}Generated documentation can be viewed at: file://{index_path}'
                ))
            else:
                self.stdout.write(self.styler.warning(
                    f'{output_prefix}Generated documentation not found, expected at: {index_path}'
                ))



import argparse
import os
import re
import subprocess
import sys
from inspect import cleandoc

from jogger.exceptions import TaskDefinitionError, TaskError
from jogger.utils.input import get_task_settings, get_tasks
from jogger.utils.output import OutputWrapper

TASK_NAME_RE = re.compile(r'^\w+$')

#
# The class-based "task" interface is heavily based on Django's management
# command infrastructure, found in ``django.core.management.base``, though
# greatly simplified and without any of the Django machinery.
#


class Task:
    """
    An advanced ``jogger`` task capable of defining its own arguments and
    redirecting the ``stdout`` and ``stderr`` output streams.
    """
    
    help = ''
    
    def __init__(self, name, settings, argv, stdout, stderr):
        
        self.settings = settings
        
        parser = self.create_parser(name, stdout, stderr)
        options = parser.parse_args(argv)
        
        kwargs = vars(options)
        
        stdout = kwargs['stdout']
        stderr = kwargs['stderr']
        no_color = kwargs['no_color']
        self.stdout = OutputWrapper(stdout, no_color=no_color)
        self.stderr = OutputWrapper(stderr, no_color=no_color, default_style='error')
        self.styler = self.stdout.styler
        
        self.using_system_out = stdout.name == '<stdout>'
        self.using_system_err = stderr.name == '<stderr>'
        
        # # TODO: Comment
        # if not self.using_system_out:
        #     self.stdout.reconfigure(write_through=True)
        #
        # if not self.using_system_err:
        #     self.stderr.reconfigure(write_through=True)
        
        self.args = kwargs.pop('args', ())
        self.kwargs = kwargs
    
    def create_parser(self, name, default_stdout, default_stderr):
        """
        Create and return the ``ArgumentParser`` which will be used to parse
        the arguments to this task.
        """
        
        parser = argparse.ArgumentParser(
            prog=name,
            description=self.help or None
        )
        
        parser.add_argument(
            '-v', '--verbosity',
            default=1,
            type=int,
            choices=[0, 1, 2, 3],
            help='Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output'
        )
        
        parser.add_argument(
            '--stdout',
            nargs='?',
            type=argparse.FileType('w'),
            default=default_stdout
        )
        
        parser.add_argument(
            '--stderr',
            nargs='?',
            type=argparse.FileType('w'),
            default=default_stderr
        )
        
        parser.add_argument(
            '--no-color',
            action='store_true',
            help="Don't colourise the command output.",
        )
        
        self.add_arguments(parser)
        
        return parser
    
    def add_arguments(self, parser):
        """
        Custom tasks should override this method to add any custom command line
        arguments they require.
        """
        
        # Do nothing - just a hook for subclasses to add custom arguments
        pass
    
    def cli(self, cmd, capture=False):
        """
        Run a command on the system's command line, in the context of the task's
        :attr:`~Task.stdout` and :attr:`~Task.stderr` output streams. Output
        can be captured rather than displayed using ``capture=True``.
        
        :param cmd: The command string to execute.
        :param capture: ``True`` to capture all output from the command rather
            than writing it to the configured output streams.
        :return: The command result object.
        """
        
        # Flush output buffers before executing the command. This is important
        # if output streams are redirected to files, to ensure any output
        # generated by the task itself is written before any output from the
        # command to be executed.
        # self.stdout.close()
        # self.stderr.close()
        # print('=====')
        # print('before:', cmd)
        # print(open('/tmp/out').read())
        # print('---')
        kwargs = {}
        if capture:
            kwargs['capture_output'] = True
        else:
            # Pass redirected output streams if necessary
            if not self.using_system_out:
                kwargs['stdout'] = self.kwargs['stdout']
            
            if not self.using_system_err:
                kwargs['stderr'] = self.kwargs['stderr']
        
        r = subprocess.run(cmd, shell=True, **kwargs)
        # print('after:', cmd)
        # print(open('/tmp/out').read())
        # print('=====')
        return r
    
    def get_task_proxy(self, task_name, *args):
        """
        Return an object representing the task matching the given name,
        configured with the given arguments, if any. This proxy object can be
        used to execute the task, regardless of whether it is defined as a
        string, function, or class::
        
            proxy = self.get_task_proxy('test')
            proxy.execute()
        
        Arguments only apply if the target task is defined as a class, and
        should be provided as individual strings, e.g.::
        
            proxy = get_task_proxy('test', '-v', '2', 'myapp.tests', '--keepdb')
        
        If the target task is defined as a class, common arguments of the
        source task will be propagated automatically, including:
        ``-v``/``--verbosity``, ``--stdout``, ``--stderr``, and ``--no-color``.
        
        :param task_name: The task name as a string.
        :param args: Extra task arguments, as individual strings.
        :return: The task proxy instance.
        """
        
        try:
            task = get_tasks()[task_name]
        except FileNotFoundError as e:
            raise TaskDefinitionError(e)
        except KeyError:
            raise TaskDefinitionError(f'Unknown task "{task_name}".')
        
        # Get the proxy instance, allow raising TaskDefinitionError if necessary
        proxy = TaskProxy('proxy.execute', task_name, task, self.stdout, self.stderr)
        
        if proxy.has_own_args:
            # The target task is also class-based, so common arguments of the
            # source task can be propagated, if not provided explicitly
            args = list(args)
            
            if '-v' not in args and '--verbosity' not in args:
                args.extend(('--verbosity', str(self.kwargs['verbosity'])))
            
            if '--no-color' not in args and self.kwargs['no_color']:
                args.append('--no-color')
            
            proxy.argv = args
        elif args:
            raise TaskError('String- and function-based tasks do not accept arguments.')
        
        return proxy
    
    def execute(self):
        """
        Execute this task. Intercept any raised ``TaskError`` and print it
        sensibly to ``stderr``. Allow all other exceptions to raise as per usual.
        """
        
        try:
            self.handle(*self.args, **self.kwargs)
        except TaskError as e:
            self.stderr.write(str(e))
            sys.exit(1)
    
    def handle(self, *args, **kwargs):
        """
        The actual logic of the task. Subclasses must implement this method.
        """
        
        raise NotImplementedError('Subclasses of Task must provide a handle() method.')


class TaskProxy:
    """
    A helper for identifying and executing tasks of different types. It will
    identify and execute the following:
    
    - Strings: Executed as-is on the command line.
    - Callables (e.g. functions): Called with ``settings``, ``stdout``, and
        ``stderr`` as keyword arguments, allowing the task to alter its
        behaviour on a per-project basis and use separate output streams if
        necessary.
    - ``Task`` class objects: Instantiated with the remainder of the argument
        string (that not consumed by the ``jog`` program itself) and executed.
        Also has access to project-level settings and the ``stdout``/``stderr``
        output streams, in addition to accepting its own custom arguments.
    """
    
    def __init__(self, prog, name, task, stdout, stderr, argv=None):
        
        try:
            valid_name = TASK_NAME_RE.match(name)
        except TypeError:  # not a string
            valid_name = False
        
        if not valid_name:
            raise TaskDefinitionError(
                f'Task name "{name}" is not valid - must be a string '
                'containing alphanumeric characters and the underscore only.'
            )
        
        if isinstance(task, str):
            self.exec_mode = 'cli'
            self.executor = self.execute_string
            self.help_text = task
            self.has_own_args = False
        elif isinstance(task, type) and issubclass(task, Task):
            self.exec_mode = 'python'
            self.executor = self.execute_class
            self.help_text = task.help
            self.has_own_args = True
        elif callable(task):
            self.exec_mode = 'python'
            self.executor = self.execute_callable
            self.help_text = cleandoc(task.__doc__) if task.__doc__ else ''
            self.has_own_args = False
        else:
            raise TaskDefinitionError(f'Unrecognised task format for "{name}".')
        
        self.prog = f'{prog} {name}'
        self.name = name
        
        self.task = task
        self.argv = argv
        
        self.stdout = stdout
        self.stderr = stderr
    
    def output_help_line(self):
        
        stdout = self.stdout
        styler = stdout.styler
        
        name = styler.heading(self.name)
        help_text = self.help_text
        if self.exec_mode == 'cli':
            help_text = styler.apply(help_text, fg='green')
        else:
            help_text = styler.apply(help_text, fg='blue')
        
        stdout.write(f'{name}: {help_text}')
        if self.has_own_args:
            stdout.write(f'    See "{self.prog} --help" for usage details')
    
    def parse_simple_args(self, help_text):
        
        parser = argparse.ArgumentParser(
            prog=self.prog,
            description=help_text,
            formatter_class=argparse.RawTextHelpFormatter
        )
        
        return parser.parse_args(self.argv)
    
    def execute(self):
        
        # Proxy to the appropriate method, sensibly handling any raised
        # TaskError (which will already be intercepted for class-based tasks,
        # but not for string or function-based ones)
        try:
            self.executor()
        except TaskError as e:
            self.stderr.write(str(e))
            sys.exit(1)
    
    def execute_string(self):
        
        help_text = f'Executes the following task on the command line:\n{self.help_text}'
        self.parse_simple_args(help_text)
        
        os.system(self.task)
    
    def execute_callable(self):
        
        self.parse_simple_args(self.help_text)
        
        settings = get_task_settings(self.name)
        self.task(settings=settings, stdout=self.stdout, stderr=self.stderr)
    
    def execute_class(self):
        
        settings = get_task_settings(self.name)
        
        # Don't pass through the OutputWrapper instances themselves, just the
        # stream they wrap. The Task instance will create its own OutputWrapper
        # around it, potentially with a different configuration (depending on
        # arguments such as --no-color). But passing through the underlying
        # streams allows reusing them, which is particularly important when
        # calling nested tasks with streams redirected to files, so they append
        # to the same file rather than overwriting each other.
        stdout = self.stdout._out
        stderr = self.stderr._out
        
        t = self.task(self.prog, settings, self.argv, stdout, stderr)
        t.execute()



import argparse
import os

from .base import Task, TaskError

try:
    import django  # noqa
    HAS_DJANGO = True
except ImportError:
    HAS_DJANGO = False

try:
    import coverage  # noqa
    HAS_COVERAGE = True
except ImportError:
    HAS_COVERAGE = False

try:
    import tblib  # noqa
    HAS_TBLIB = True
except ImportError:
    HAS_TBLIB = False


class TestTask(Task):
    
    help = (
        'Run the test suite. If coverage.py is detected, perform code '
        'coverage analysis, print an on-screen summary, and generate a fully '
        'detailed HTML report.'
    )
    
    def __init__(self, *args, **kwargs):
        
        self._has_output = False
        
        super().__init__(*args, **kwargs)
    
    def add_arguments(self, parser):
        
        parser.add_argument(
            'paths',
            nargs='*',
            metavar='test_path',
            help='Module paths to test.'
        )
        
        parser.add_argument(
            '-q', '--quick',
            action='store_true',
            help=(
                'Run a "quick" variant of the task: no coverage analysis and '
                'running tests in parallel (where possible).'
            )
        )
        
        parser.add_argument(
            '-a',
            action='store_true',
            dest='accumulate',
            help=(
                'Accumulate coverage data across multiple runs. Disables all '
                'coverage reporting (to be run after all coverage data is'
                'accumulated).'
            )
        )
        
        parser.add_argument(
            '--report',
            action='store_true',
            dest='reports_only',
            help=(
                'Skip the test suite and just output the on-screen summary '
                'and generate the HTML report. Useful to review previous '
                'results or if using -a to accumulate results before running '
                'the reports.'
            )
        )
        
        parser.add_argument(
            '--no-html',
            action='store_false',
            dest='html_report',
            help='Do not generate a detailed HTML code coverage report.'
        )
        
        parser.add_argument(
            '--src',
            default='',
            dest='source',
            help=(
                'Specify the source to measure the coverage of.'
            )
        )
        
        parser.add_argument('extra', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)
    
    @property
    def section_prefix(self):
        
        # If the command has generated any output, sections should be prefixed
        # with a newline. The first section to generate output should not.
        if self._has_output:
            return '\n'
        
        return ''
    
    def get_coverage_command(self, quick, source, accumulate, **options):
        
        if not HAS_COVERAGE or quick:
            return ''
        
        if source:
            source = f' --source {source}'
        
        accumulate = ' -a' if accumulate else ''
        
        return f'coverage run --branch{source}{accumulate} '
    
    def get_test_command(self, labels, coverage, quick, verbosity, extra, **options):
        
        command = []
        
        if not coverage:
            # Run with warnings enabled, unless the command will be run by
            # coverage (which expects a python script to run, not the `python`
            # program)
            command.append('python -Wa')
        
        command.append('manage.py test')
        command.extend(labels)
        
        # Pass all "extra" arguments, and the verbosity level, through to the
        # test command
        command.extend(extra)
        command.append(f'-v {verbosity}')
        
        # Add --parallel switch if using "quick" mode, unless disabled in
        # settings or if the switch is provided explicitly in "extra" arguments
        if not any(v.startswith('--parallel') for v in extra):
            parallel = self.settings.get('quick_parallel', None)
            if quick and (parallel is None or int(parallel) > 1):
                if not HAS_TBLIB:
                    self.stdout.write(self.styler.warning(
                        'Tracebacks in parallel tests may not display correctly: '
                        'tblib not detected. pip install tblib to fix.'
                    ))
                
                command.append('--parallel')
                if parallel:
                    command.append(parallel)
        
        return ' '.join(command)
    
    def do_summary(self, verbosity, **options):
        
        if verbosity < 1:
            return
        
        self.stdout.write(self.styler.label(f'{self.section_prefix}Coverage summary'))
        if verbosity > 1:
            self.cli('coverage report')
        else:
            self.cli('coverage report --skip-covered')
    
    def do_html_report(self, html_report, **options):
        
        if not html_report:
            return
        
        self.stdout.write(self.styler.label(f'{self.section_prefix}Generating HTML report...'), ending=None)
        self.cli('coverage html')
        
        html_report_path = os.path.abspath('htmlcov/index.html')
        if os.path.exists(html_report_path):
            html_report_path = f'file://{html_report_path}'
        else:
            html_report_path = f'Location of HTML report unknown, expected {html_report_path}'
        
        self.stdout.write(f' done: {html_report_path}')
    
    def handle(self, *args, **options):
        
        reports_only = options['reports_only']
        if reports_only:
            if options['quick']:
                raise TaskError('-q/--quick and --report are mutually exclusive.')
            elif options['accumulate']:
                raise TaskError('-a and --report are mutually exclusive.')
        
        if not HAS_DJANGO:
            raise TaskError('Django not detected.')
        
        if not reports_only:
            test_paths = options.pop('paths', None)
            coverage_command = self.get_coverage_command(**options)
            test_command = self.get_test_command(test_paths, coverage=bool(coverage_command), **options)
            
            self.cli(f'{coverage_command}{test_command}')
        
        if not HAS_COVERAGE:
            # Not having coverage available is simply a warning unless directly
            # requesting coverage reports, in which case it is an error
            msg = 'Code coverage not available: coverage.py not detected'
            if not reports_only:
                self.stdout.write(self.styler.warning(msg))
            else:
                raise TaskError(msg)
        elif not options['accumulate'] and not options['quick']:
            self.do_summary(**options)
            self.do_html_report(**options)
    
    def cli(self, cmd):
        
        self._has_output = True
        
        return super().cli(cmd)



import configparser
import os
from importlib.util import module_from_spec, spec_from_file_location

from jogger.exceptions import TaskDefinitionError

from .files import find_file

MAX_CONFIG_FILE_SEARCH_DEPTH = 8
JOG_FILE_NAME = 'jog.py'
CONFIG_FILE_NAME = 'setup.cfg'
CONFIG_BLOCK_PREFIX = 'jogger'


def find_config_file(target_file_name):
    """
    Search upwards from the current working directory looking for a config file
    named ``target_file_name``. The number of parent directories to search
    through is controlled by ``MAX_CONFIG_FILE_SEARCH_DEPTH``.
    
    :param target_file_name: The filename of the target file.
    :return: The absolute path of the located file.
    """
    
    path = os.getcwd()
    
    return find_file(target_file_name, path, MAX_CONFIG_FILE_SEARCH_DEPTH)


def get_tasks():
    """
    Search upwards from the current working directory looking for the task
    definition file (``JOG_FILE_NAME``). Import the file as a Python module
    and return its inner ``tasks`` dictionary. Raise ``TaskDefinitionError``
    if no ``tasks`` dictionary is defined in the imported module.
    
    :return: The task definition file's dictionary of tasks.
    """
    
    jog_file = find_config_file(JOG_FILE_NAME)
    
    spec = spec_from_file_location('jog', jog_file)
    jog_file = module_from_spec(spec)
    spec.loader.exec_module(jog_file)
    
    try:
        return jog_file.tasks
    except AttributeError:
        raise TaskDefinitionError(f'No tasks dictionary defined in {JOG_FILE_NAME}.')


def get_task_settings(task_name):
    """
    Search upwards from the current working directory looking for the project
    config file (``CONFIG_FILE_NAME``). Parse the file and return the section
    corresponding to ``task_name``. If no such section exists, return an empty
    section.
    
    :return: The config file section, as a ``configparser.SectionProxy`` object,
        for the given task.
    """
    
    config_file = configparser.ConfigParser()
    
    try:
        config_file_path = find_config_file(CONFIG_FILE_NAME)
    except FileNotFoundError:
        # Silently ignore non-existent settings files, they are not mandatory
        pass
    else:
        config_file.read(config_file_path)
    
    section = f'{CONFIG_BLOCK_PREFIX}:{task_name}'
    
    # If the section does not exist, add a dummy one. This allows this method
    # to always return a value of a consistent type.
    if not config_file.has_section(section):
        config_file.add_section(section)
    
    return config_file[section]



import argparse
import sys

from jogger import __version__ as version
from jogger.exceptions import TaskDefinitionError
from jogger.tasks.base import TaskProxy
from jogger.utils.input import JOG_FILE_NAME, get_tasks
from jogger.utils.output import OutputWrapper


def parse_args(prog, argv=None):
    
    parser = argparse.ArgumentParser(
        prog=prog,
        formatter_class=argparse.RawTextHelpFormatter,
        description='Execute common, project-specific tasks.',
        epilog=(
            'Any additional arguments are passed through to the executed tasks.'
            '\n\n'
            'Run without arguments from within a target project to output all '
            f'tasks configured in that project\'s {JOG_FILE_NAME} file.'
        )
    )
    
    parser.add_argument(
        'task_name',
        nargs='?',
        metavar='task',
        help='The name of the task'
    )
    
    parser.add_argument(
        '--version',
        action='version',
        version=f'jogger {version}',
        help='Display the version number and exit'
    )
    
    parser.add_argument('extra', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)
    
    return parser.parse_args(argv)


def main(argv=None):
    
    prog = 'jog'
    arguments = parse_args(prog, argv)
    stdout = OutputWrapper(sys.stdout)
    stderr = OutputWrapper(sys.stderr, default_style='error')
    
    try:
        tasks = get_tasks()
        for name, task in tasks.items():
            tasks[name] = TaskProxy(prog, name, task, stdout, stderr, arguments.extra)
    except (FileNotFoundError, TaskDefinitionError) as e:
        stderr.write(str(e))
        sys.exit(1)
    
    task_name = arguments.task_name
    if task_name:
        try:
            task = tasks[task_name]
        except KeyError:
            stderr.write(f'Unknown task "{task_name}".')
            sys.exit(1)
        
        task.execute()
    elif not tasks:
        stdout.write('No tasks defined.')
    else:
        stdout.write('Available tasks:', 'label')
        for task in tasks.values():
            task.output_help_line()


if __name__ == '__main__':
    main()



import os
import sys
from io import TextIOBase

#
# This module is heavily based on Django, adapted from functionality found in
# ``django.core.management.base``, ``django.core.management.color``, and
# ``django.utils.termcolors``.
#

COLOR_NAMES = ('black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white')
FOREGROUND = {COLOR_NAMES[x]: '3%s' % x for x in range(8)}
BACKGROUND = {COLOR_NAMES[x]: '4%s' % x for x in range(8)}

OPTIONS = {'bold': '1', 'underscore': '4', 'blink': '5', 'reverse': '7', 'conceal': '8'}
RESET = '\x1b[0m'


class Styler:
    """
    An object containing methods for generating styled text for a palette of
    preconfigured style roles. Text is styled by wrapping it in appropriate
    ANSI graphics codes.
    
    Text styling can be disabled using the ``no_color`` constructor argument.
    When ``True``, all methods will return the provided text unmodified. This
    enables a common API between environments that support styled text and
    those that do not.
    
    :attr:`~Styler.PALETTE` defines the name and attributes of each
    preconfigured role. Its entries are mapped into methods on the class that
    can be used as shortcuts to apply the corresponding set of style attributes
    to the given text. Subclasses can define their own palettes.
    
    Additionally, the :meth:`~Styler.apply` method can be used to apply any
    arbitrary text styles if the :attr:`~Styler.PALETTE` configuration doesn't
    include a suitable role.
    
    Usage::
    
        styler = Styler()
        success_message = styler.success('It worked!')
        error_message = styler.error('It failed!')
        message = styler.apply('hello', fg='red', bg='blue', opts=('blink', ))
    """
    
    PALETTE = {
        'success': {'fg': 'green', 'options': ('bold', )},
        'error': {'fg': 'red', 'options': ('bold', )},
        'warning': {'fg': 'yellow', 'options': ('bold', )},
        'info': {'options': ('bold', )},
        'debug': {'fg': 'magenta', 'options': ('bold', )},
        'heading': {'fg': 'cyan', 'options': ('bold', )},
        'label': {'options': ('bold', )},
    }
    
    def __init__(self, no_color=False):
        
        self.no_color = no_color
        
        for role, fmt in self.PALETTE.items():
            setattr(self, role, self.preconfigure(**fmt))
    
    def preconfigure(self, **kwargs):
        """
        Return a function with default parameters for ``apply()``.
        
        Examples::
            
            bold_red = styler.preconfigure(opts=('bold',), fg='red')
            print(bold_red('hello'))
            
            KEYWORD = styler.preconfigure(fg='yellow')
            COMMENT = styler.preconfigure(fg='blue', opts=('bold',))
        """
        
        return lambda text: self.apply(text, **kwargs)
    
    def apply(self, text, fg=None, bg=None, options=(), reset=True):
        """
        Return ``text``, enclosed in ANSI graphics codes, as dictated by
        ``fg``, ``bg``, and ``options``. If ``reset`` is ``True``, the returned
        text will be terminated by the RESET code.
        
        If configured with ``no_color=True``, return the text unmodified.
        
        Valid colors::
        
            'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'
        
        Valid options::
        
            'bold', 'underscore', 'blink', 'reverse', 'conceal'
        
        Examples::
            
            styler.apply('hello', fg='red', bg='blue', opts=('blink', ))
            styler.apply('goodbye', opts=('underscore', ))
            print(styler.apply('first line', fg='red', reset=False))
            print('this should be red too')
            print(styler.apply('and so should this'))
            print('this should not be red')
        
        :param text: The text to style.
        :param fg: The foreground colour to apply to ``text``.
        :param bg: The background colour to apply to ``text``.
        :param options: The display options to apply to ``text``.
        :param reset: ``True`` to clear all applied styles at the end of ``text``,
            ``False`` to allow them to affect subsequent output.
        :return: The styled text.
        """
        
        if self.no_color:
            return text
        
        if reset:
            text = f'{text}{RESET}'
        
        code_list = []
        
        if fg:
            code_list.append(FOREGROUND[fg])
        
        if bg:
            code_list.append(BACKGROUND[bg])
        
        for o in options:
            code_list.append(OPTIONS[o])
        
        if code_list:
            code_list = ';'.join(code_list)
            text = f'\x1b[{code_list}m{text}'
        
        return text
    
    def reset(self):
        """
        Return the ANSI RESET graphics code. Can be used to reset a style
        created by calling ``apply(..., reset=False)``.
        
        If configured with ``no_color=True``, return an empty string.
        
        :return: The ANSI RESET graphics code.
        """
        
        if self.no_color:
            return ''
        
        return RESET


class OutputWrapper(TextIOBase):
    """
    Simple wrapper around ``stdout``/``stderr`` to normalise some behaviours.
    """
    
    def __init__(self, out, default_style=None, no_color=False):
        
        self._out = out
        
        no_color = no_color or not self.supports_color()
        self.styler = Styler(no_color)
        self.default_style = default_style
    
    def __getattr__(self, name):
        
        return getattr(self._out, name)
    
    def supports_color(self):
        """
        Return True if the output stream supports color, and False otherwise.
        """
        
        plat = sys.platform
        supported_platform = plat != 'Pocket PC' and (plat != 'win32' or 'ANSICON' in os.environ)
        is_a_tty = hasattr(self._out, 'isatty') and self._out.isatty()
        
        return supported_platform and is_a_tty
    
    def write(self, msg, style=None, ending='\n'):
        
        if ending:
            msg += ending
        
        style = style or self.default_style
        if style:
            msg = getattr(self.styler, style)(msg)
        
        self._out.write(msg)



import os
from fnmatch import fnmatch as std_fnmatch


def find_file(target_file_name, from_path, max_search_depth=16):
    """
    Search upwards from ``from_path`` looking for ``target_file_name``,
    through a maximum of ``max_search_depth`` parent directories. Raise
    ``FileNotFoundError`` if the target file is not located.
    
    :param target_file_name: The filename of the target file.
    :param from_path: The directory in which to begin the search.
    :param max_search_depth: The maximum number of parent directories to search
        up through.
    :return: The absolute path of the located file.
    """
    
    path = from_path
    matched_file = None
    depth = 0
    
    while path and depth < max_search_depth:
        filename = os.path.join(path, target_file_name)
        
        if os.path.exists(filename):
            matched_file = filename
            break
        
        new_path = os.path.dirname(path)
        if new_path == path:
            break
        
        path = new_path
        depth += 1
    
    if not matched_file:
        raise FileNotFoundError(f'Could not find {target_file_name}.')
    
    return matched_file


def fnmatch(filename, patterns):
    """
    Test whether the ``filename`` string matches any of the strings in
    ``patterns``, using the standard library ``fnmatch()`` function internally
    to test each pattern. Return ``True`` if the filename matches or ``False``
    if it does not.
    
    :param filename: The filename to test.
    :param patterns: An iterable of patterns to test against.
    :return: ``True`` if a match is found, ``False`` if not.
    """
    
    return any(std_fnmatch(filename, pattern) for pattern in patterns)


def pathmatch(path, patterns):
    """
    Test whether the ``path`` string matches any of the strings in ``patterns``.
    ``path`` can be a relative file path, with both its basename and absolute
    path also tested against ``patterns``.
    
    :param path: The file path to test.
    :param patterns: An iterable of patterns to test against.
    :return: ``True`` if a match is found, ``False`` if not.
    """
    
    if fnmatch(path, patterns):
        return True
    
    basename = os.path.basename(path)
    if fnmatch(basename, patterns):
        return True
    
    absolute_path = os.path.abspath(path)
    return fnmatch(absolute_path, patterns)


def walk(from_path, exclude_patterns=None):
    """
    Yield all filenames under the ``from_path`` directory, optionally excluding
    any files/directories matching any of the pattern strings in
    ``exclude_patterns``.
    
    :param from_path: The root directory to walk.
    :param exclude_patterns: An iterable of patterns to test against.
    """
    
    if not exclude_patterns:
        # No exclusion patterns, perform simple directory walk
        for root, dirs, files in os.walk(from_path):
            for filename in files:
                yield os.path.join(root, filename)
    else:
        # Perform more complex directory walk, excluding files/directories
        # matching given exclusion patterns
        for root, dirs, files in os.walk(from_path, topdown=True):
            # Removing items from `dirs` will prevent `os.walk` from entering
            # those subdirectories. Iterate a copy so removals don't result in
            # an early exit from the loop.
            for directory in dirs.copy():
                joined = os.path.join(root, directory)
                if pathmatch(joined, exclude_patterns):
                    dirs.remove(directory)
            
            for filename in files:
                joined = os.path.join(root, filename)
                if not pathmatch(joined, exclude_patterns):
                    yield joined



from jogger.exceptions import TaskError  # noqa - for convenience

from .base import Task  # noqa
from .docs import DocsTask  # noqa
from .lint import LintTask  # noqa
from .test import TestTask  # noqa
from .update import UpdateTask  # noqa



import os
import shutil
import sys

from jogger.utils.input import JOG_FILE_NAME, find_config_file

from .base import Task, TaskDefinitionError, TaskError


class UpdateTask(Task):
    
    help = (
        'Update the application by checking for remote changes, pulling them '
        'if found, updating dependencies if necessary, migrating the database, '
        'collecting static files, and restarting the relevant services.'
    )
    
    temp_requirements_dir = '/tmp'
    
    def add_arguments(self, parser):
        
        parser.add_argument(
            '--no-input',
            action='store_true',
            dest='no_input',
            help=(
                'Do not prompt for user input, e.g. to confirm dependency '
                'updates or migrations.'
            )
        )
    
    def handle(self, **options):
        
        self.check_updates()
        
        requirements_path, temp_requirements_path = self.check_initial_requirements()
        
        self.do_pull()
        self.do_dependency_check(requirements_path, temp_requirements_path)
        self.do_migration_check()
        self.do_stale_contenttypes_check()
        self.do_build()
        self.do_collect_static()
        self.do_restart()
        
        self.stdout.write('\nDone!', style='label')
    
    def check_updates(self):
        
        self.stdout.write('Checking for updates', style='label')
        
        result = self.cli('git log --oneline origin master..master | wc -l', capture=False)
        
        if result.returncode:
            sys.exit(1)
        
        update_count = int(result.stdout)
        if not update_count:
            self.stdout.write('No remote changes')
            sys.exit(0)
        
        self.stdout.write(f'Found {update_count} new remote commits')
    
    def check_initial_requirements(self):
        
        # Get the project directory by locating the task file (jog.py)
        project_dir = os.path.dirname(find_config_file(JOG_FILE_NAME))
        project_name = os.path.split(project_dir)[1].replace('-', '_')
        
        requirements_path = os.path.join(project_dir, 'requirements.txt')
        temp_requirements_path = os.path.join(
            self.temp_requirements_dir,
            f'{project_name}.lastpull.requirements.txt'
        )
        
        # If this is the first time the command has been run, make a copy of
        # the requirements.txt file prior to pulling
        if not os.path.exists(temp_requirements_path):
            shutil.copyfile(requirements_path, temp_requirements_path)
        
        return requirements_path, temp_requirements_path
    
    def do_pull(self):
        
        self.stdout.write('\nPulling', style='label')
        
        result = self.cli('git pull origin master --prune --no-rebase')
        
        if result.returncode:
            # Stop script here if the pull was not successful for any reason
            raise TaskError('Pull errored/aborted')
    
    def do_dependency_check(self, requirements_path, temp_requirements_path):
        
        self.stdout.write('\nChecking Python library dependencies', style='label')
        
        # Check for dependency updates by diffing the stored requirements.txt
        # file with the one just pulled in
        diff_result = self.cli(f'diff -U 0 {temp_requirements_path} {requirements_path}', capture=True)
        
        if not diff_result.returncode:
            self.stdout.write('No changes detected')
            return
        
        # Changes were detected, show the differences and prompt the user
        # whether to proceed with an install or not. Alternatively, if running
        # in no-input mode, proceed directly to the install.
        if self.kwargs['no_input']:
            answer = 'y'
        else:
            self.stdout.write(diff_result.stdout.decode('utf-8'))
            
            answer = input(
                'The above Python library dependency changes were detected, '
                'update now (Y/n)? '
            )
        
        if answer.lower() == 'y':
            install_result = self.cli(f'pip install -r {requirements_path}')
            if not install_result.returncode:
                # Make a copy of the now-applied requirements.txt to compare
                # next time the task is run
                shutil.copy(requirements_path, temp_requirements_path)
        elif answer.lower() == 'n':
            self.stdout.write('Dependency update skipped')
        else:
            # User didn't answer yes OR no, display an error message but
            # don't interrupt execution
            self.stdout.write('Dependency update aborted', style='error')
    
    def do_migration_check(self):
        
        self.stdout.write('\nChecking migrations', style='label')
        
        # Django doesn't give a nice command to find unapplied migrations, so
        # grep the entire migration list to exclude:
        # - applied migrations (denoted by [X])
        # - (no migrations) lines for apps without migrations
        # - leftover app names with no migrations listed between them and
        #   either the next app or the end of the input
        # The grep -v flag excludes the lines that match and the -P flag is
        # used to activate perl-regexp, which enables using \n.
        # W605 (invalid escape sequence) is ignored because they *are* valid
        # escape sequences, just not used in a Python-based regex operation.
        cmd = (
            "python manage.py showmigrations --list "
            "| grep -v '\[X\]' "  # noqa W605
            "| grep -v 'no migrations' "
            "| grep -Pv '^[a-zA-Z0-9_]+(?=\Z|\n[a-zA-Z_])'"  # noqa W605
        )
        list_result = self.cli(cmd, capture=True)
        
        if list_result.returncode:
            raise TaskError('Migration check failed')
        
        list_output = list_result.stdout.decode('utf-8')
        if list_output:
            self.stdout.write('No changes detected')
            return
        
        # Changes were detected, show them and prompt the user whether to
        # proceed with a migration or not. Alternatively, if running in
        # no-input mode, proceed directly with the migrations.
        if self.kwargs['no_input']:
            answer = 'y'
        else:
            self.stdout.write(list_output)
            answer = input('The above migrations are unapplied, apply them now (Y/n)? ')
        
        if answer.lower() == 'y':
            self.cli('python manage.py migrate')
        elif answer.lower() == 'n':
            self.stdout.write('Migrations skipped')
        else:
            # User didn't answer yes OR no, display an error message but
            # don't interrupt execution
            self.stdout.write('Migrations aborted', style='error')
    
    def do_stale_contenttypes_check(self):
        
        if self.kwargs['no_input']:
            # Due to the possibility of deleting records, do not remove stale
            # content types when running in no-input mode
            return
        
        self.stdout.write('\nChecking stale content types', style='label')
        
        self.cli('python manage.py remove_stale_contenttypes')
    
    def do_build(self):
        
        try:
            proxy = self.get_task_proxy('build')
        except TaskDefinitionError:
            # A "build" task either isn't defined or is invalidly defined.
            # Do nothing.
            pass
        else:
            self.stdout.write('\nRunning build/s', style='label')
            proxy.execute()
    
    def do_collect_static(self):
        
        self.stdout.write('\nCollecting static files', style='label')
        
        cmd = 'python manage.py collectstatic'
        if self.kwargs['no_input']:
            cmd = f'{cmd} --no-input'
        
        self.cli(cmd)
    
    def do_restart(self):
        
        # Hook for subclasses to restart the necessary services after the
        # update has been completed
        pass
